/* DSP LABORATORY 
 * University of Technology Graz, 2018
 * Josef Kulmer
 * 
 * Fouriertransform calculates the spectrum of a provided input signal. The 
 * input signal is multiplied by a hamming window. The outcome is averaged and
 * forwarded to the output.
 * */

#include <fftw3.h>
#include "math.h"
#include <random>
#include <iostream>
#include "blockprocessing.h"
#include "Fouriertransform.h"

Fouriertransform::Fouriertransform() { //constructor
    // initialize fft operations
    inft = (float*) fftwf_malloc(sizeof(float) * FRAMELENGTH);
    outft = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * FRAMELENGTH);
    
    outift = (float*) fftwf_malloc(sizeof(float) * FRAMELENGTH);
    inift = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * FRAMELENGTH);
    
    p = fftwf_plan_dft_r2c_1d(FRAMELENGTH, inft, outft, FFTW_ESTIMATE);
    pi = fftwf_plan_dft_c2r_1d(FRAMELENGTH, inift, outift, FFTW_ESTIMATE);
    
    // initialize hamming analysis window
    windowfcn.resize(FRAMELENGTH);
    for(unsigned int i=0;i<FRAMELENGTH;i++){
        windowfcn[i]=0.54-0.46*cos(2*PI*i/(FRAMELENGTH-1)); // hamming window equation
    }
};

Fouriertransform::~Fouriertransform() { //deconstructor
    fftwf_destroy_plan(p);
    fftwf_destroy_plan(pi);
    fftwf_free(outft);
    fftwf_free(inft);
    fftwf_free(outift);
    fftwf_free(inift);
};

void Fouriertransform::doWindowfcn(const void *in, unsigned long framelength) {
    // apply a window function to the input
    float *input = (float*) in;
    for(int i=0;i<framelength;i++){
        input[i] = input[i] * windowfcn[i];
    }
}

void Fouriertransform::doDFT(const void *in, const void *outreal, const void *outimag, unsigned long framelength) {
// performs the Fouriertransform
    float *input = (float*) in;
    float *outputr = (float*) outreal;
    float *outputi = (float*) outimag;
    
    /* copy input data to input variable
     * of the Fourier transform processing */
    for(int i=0;i<framelength;i++){
        inft[i] = *input++;
    }

    // perform Fourier transform
    fftwf_execute(p);
    
    // forward to output, separated in real and imaginary parts
    for(int i=0;i<=framelength/2;i++){ // length of spectrum of real valued dft is N/2+1
        *outputr++ = outft[i][0];
        *outputi++ = outft[i][1];
    }
}

void Fouriertransform::doIDFT(const void *inreal, const void *inimag, const void *out, unsigned long framelength) {
// performs the inverse Fouriertransform
    float *output = (float*) out;
    float *inputr = (float*) inreal;
    float *inputi = (float*) inimag;
    
    /* copy spectrum to input variable
     * of the Fourier transform processing */
    
    for(int i=0;i<=framelength/2;i++){// length of spectrum of real valued dft is N/2+1
        inift[i][0]=*inputr++;
        inift[i][1]=*inputi++;
    }
    
    fftwf_execute(pi);
    
    // provide output to output
    for(int i=0;i<framelength;i++){
        *output++ = outift[i]/framelength; // consider normalization
    }
}
