/* DSP LABORATORY 
 * University of Technology Graz, 2018
 * Josef Kulmer
 * 
 * UNIT 5: Speech Enhancement
 * 
 * 
 */
#include <iostream>
#include <math.h>
#include "portaudio.h"
#include "blockprocessing.h"
#include "Fouriertransform.h"
#include "WienerFilter.h"

// collect objects for computation in this struct
struct UserDataExperiment{
    int checkCallback;
    int counter;
    int currentstate;
    float specreal[FRAMELENGTH/2+1];
    float specimag[FRAMELENGTH/2+1];
    // left input channel (signal plus noise)
    float buf0l[FRAMELENGTH]; // present frame, unprocessed
    float buf0pl[FRAMELENGTH]; // present frame, processed
    float buf1l[FRAMELENGTH]; // present frame-1, unprocessed
    float buf1pl[FRAMELENGTH]; // present frame-1, processed
    float buf2outl[FRAMELENGTH/2]; // result present frame-2
    float buf2inl[FRAMELENGTH/2]; // unprocessed present frame-2
    // right input channel (noise)
    float buf0r[FRAMELENGTH]; // present frame
    float buf1r[FRAMELENGTH]; // present frame-1
    float buf2inr[FRAMELENGTH/2]; // unprocessed present frame-2
    
    WienerFilter WF;
};

// the callback function, evaluated at each interrupt
static int paCallback(const void *input, void *output,
    unsigned long framelength,
    const PaStreamCallbackTimeInfo* timeInfo,
    PaStreamCallbackFlags statusFlags,
    void *userdata)
{
    float *in = (float*)input;
    float *out = (float*)output;
    
    UserDataExperiment *ud = (UserDataExperiment*)userdata;
    
    // analysis starts here
    // store this frame input
    for(int i=0;i<framelength;i++){
        ud->buf0l[i]=in[i*2];
        ud->buf0r[i]=in[i*2+1];
    }
    
    // assemble overlapping buffer
    for(int i=0;i<framelength/2;i++){
        ud->buf1l[i]=ud->buf2inl[i];
        ud->buf1r[i]=ud->buf2inr[i];
        
        ud->buf1l[i+FRAMELENGTH/2]=in[i*2];
        ud->buf1r[i+FRAMELENGTH/2]=in[i*2+1];
        
        ud->buf2inl[i]=in[i*2+framelength];
        ud->buf2inr[i]=in[i*2+1+framelength];
    }
    
    // apply Speech Enhancement here
    if(ud->currentstate == 0){ // switch Speech Enhancement on/off each 2 seconds
        for(int i=0;i<framelength;i++){ // speech enhancement off
            ud->buf1pl[i]=ud->buf1l[i];
            ud->buf0pl[i]=ud->buf0l[i];
        }
    }
    if(ud->currentstate == 1) { // speech enhancement on
        ud->WF.doSpeechEnhancement(ud->buf1l,ud->buf1r,ud->buf1pl,framelength);
        ud->WF.doSpeechEnhancement(ud->buf0l,ud->buf0r,ud->buf0pl,framelength);
    }
    
    if((ud->currentstate==0) && (ud->counter * FRAMELENGTH / FS >= 2)){
        ud->currentstate = 1;
        std::cout<<"Filter on"<<std::endl;
    }
    if((ud->currentstate==1) && (ud->counter * FRAMELENGTH / FS >= 4)){
        ud->currentstate = 0;
        ud->counter = 0;
        std::cout<<"Filter off"<<std::endl;
    }

    // OLA
    // apply window function, required for synthesis
    ud->WF.FT.doWindowfcn(output,framelength);
    // forward output, lower half
    for(int i=0;i<framelength/2;i++)
        *(out+i)=(ud->buf1pl[i]+ud->buf2outl[i])/2; // divide by two, since overlap of two
    // forward output, upper half
    for(int i=0;i<framelength/2;i++)
        *(out+i+framelength/2)=(ud->buf1pl[i+FRAMELENGTH/2]+ud->buf0pl[i])/2;
    
    // store upper half of present frame (buf0) in buf2
    for(int i=0;i<framelength/2;i++)
        ud->buf2outl[i]=ud->buf0pl[i+FRAMELENGTH/2];
    
    ud->counter++;
    ud->checkCallback ++; // report that callback function is working
    return 0;
}

int main(int argc, char *argv[]) {   
    PaStream *stream;
    PaError err;
    int checkCallback_local = 0; // check whether callback function is still active
    int iTimecounter = 0;
    
    UserDataExperiment ud; // create objects
    ud.checkCallback = 0; // a watchdog if callback is still working (necessary for RPi)
    for(int i=0;i<FRAMELENGTH/2;i++){
        ud.buf2inl[i]=0;
        ud.buf2inr[i]=0;
    }
    ud.counter=0;
    ud.currentstate=0;
    // the following loop stops and restarts the portaudio stream if it is broken
    do{
        err = Pa_Initialize();
        std::cout << "Initialize portaudio: " << Pa_GetErrorText( err ) << std::endl;

        err = Pa_OpenDefaultStream(&stream, 
            NINPUTCHANNELS, 
            NOUTPUTCHANNELS, 
            paFloat32, 
            FS, 
            FRAMELENGTH, 
            paCallback, 
            &ud);
        std::cout << "Open audio stream: " << Pa_GetErrorText( err ) << std::endl;

        err = Pa_StartStream(stream);
        std::cout << "Start audio stream: " << Pa_GetErrorText( err ) << std::endl;

        Pa_Sleep(TIMETOSLEEP); // wait until Stream is open
        
        // wait in this loop until DURATION is reached or callback function needs to be restarted
        while(iTimecounter < DURATION-1 && ud.checkCallback != checkCallback_local){ // there was a callback within last second
            checkCallback_local = ud.checkCallback;
            Pa_Sleep(TIMETOSLEEP); // wait a second
            iTimecounter++;
        }
        
        if(iTimecounter != DURATION-1 && ud.checkCallback != checkCallback_local){
            std::cout << "There is a problem with portaudio -> restart" << std::endl;
        }

        err = Pa_StopStream(stream);
        std::cout << "Stop audio stream: " << Pa_GetErrorText( err ) << std::endl;

        err = Pa_CloseStream(stream);
        std::cout << "Close audio stream: " << Pa_GetErrorText( err ) << std::endl;
    }while(iTimecounter < DURATION-1);

    Pa_Terminate();
    return 0;    
}
